import json
import sqlite_utils
import yaml

COLUMNS = {
    "table": str,
    "key": str,
    "title": str,
    "timestamp": str,
    "category": int,
    "is_public": int,
    "search_1": str,
    "search_2": str,
    "search_3": str,
}
INDEXES = [("timestamp",), ("category",), ("is_public",)]
FOREIGN_KEYS = [("category", "categories", "id")]
DEFAULTS = {"is_public": 0}
NOT_NULL = {
    "is_public",
}

CATEGORIES = [
    {"id": 1, "name": "created"},
    {"id": 2, "name": "saved"},
    {"id": 3, "name": "received"},
]


def run_indexer(db_path, rules, tokenize="porter"):
    db = sqlite_utils.Database(db_path)
    ensure_table_and_indexes(db, tokenize)
    db.conn.close()

    # We connect to each database in turn and attach our index
    for i, (db_name, table_rules) in enumerate(rules.items()):
        other_db = sqlite_utils.Database(db_name)
        other_db.conn.execute("ATTACH DATABASE '{}' AS index1".format(db_path))
        for table, info in table_rules.items():
            # Execute SQL with limit 0 to figure out the columns
            sql = info["sql"]
            # Bit of a hack - we replace the starting `select ` with one
            # that also includes the hard-coded table
            sql_rest = sql.split("select", 1)[1]
            sql = "select '{}/{}' as [table],{}".format(db_name, table, sql_rest)
            columns = derive_columns(other_db, sql)
            with other_db.conn:
                other_db.conn.execute(
                    "REPLACE INTO index1.search_index ({}) {}".format(
                        ", ".join("[{}]".format(column) for column in columns), sql
                    )
                )
        other_db.conn.close()

    # Run optimize
    db = sqlite_utils.Database(db_path)
    with db.conn:
        db["search_index"].optimize()
    db.vacuum()


def derive_columns(db, sql):
    cursor = db.conn.execute(sql + " limit 0")
    return [r[0] for r in cursor.description]


def ensure_table_and_indexes(db, tokenize):
    db["categories"].insert_all(CATEGORIES, pk="id", replace=True)
    table = db["search_index"]
    if not table.exists():
        table.create(
            COLUMNS,
            pk=("table", "key"),
            not_null=NOT_NULL,
            defaults=DEFAULTS,
        )
    else:
        # Ensure all the column exists
        existing_columns = table.columns_dict.keys()
        for key, type_ in COLUMNS.items():
            if key not in existing_columns:
                table.add_column(key, type_, not_null_default=DEFAULTS.get(key))
    if not db["search_index_fts"].exists():
        table.enable_fts(["title", "search_1"], create_triggers=True, tokenize=tokenize)
    for index in INDEXES:
        table.create_index(index, if_not_exists=True)
    for fk in FOREIGN_KEYS:
        try:
            table.add_foreign_key(*fk)
        except sqlite_utils.db.AlterError:
            pass


class BadMetadataError(Exception):
    pass


def parse_metadata(content):
    # content can be JSON or YAML
    try:
        return json.loads(content)
    except json.JSONDecodeError:
        try:
            return yaml.safe_load(content)
        except yaml.YAMLError:
            raise BadMetadataError("Metadata is not valid JSON or YAML")



import click
from .utils import parse_metadata, run_indexer


@click.group()
@click.version_option()
def cli():
    "Dogsheep search index"


@cli.command()
@click.argument(
    "db_path",
    type=click.Path(file_okay=True, dir_okay=False, allow_dash=False),
    required=True,
)
@click.argument(
    "config",
    type=click.Path(file_okay=True, dir_okay=False, allow_dash=True),
    required=True,
)
@click.option(
    "--tokenize",
    help="Tokenizer to use. Defaults to porter, set to none to disable.",
    default="porter",
)
def index(db_path, config, tokenize):
    "Create a search index based on rules in the config file"
    rules = parse_metadata(open(config).read())
    run_indexer(db_path, rules, tokenize=None if tokenize == "none" else tokenize)



from setuptools import setup
import os

VERSION = "0.4"


def get_long_description():
    with open(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "README.md"),
        encoding="utf8",
    ) as fp:
        return fp.read()


setup(
    name="dogsheep-beta",
    description="Dogsheep search index",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    author="Simon Willison",
    url="https://github.com/dogsheep/beta",
    project_urls={
        "Issues": "https://github.com/dogsheep/beta/issues",
        "CI": "https://github.com/dogsheep/beta/actions",
        "Changelog": "https://github.com/dogsheep/beta/releases",
    },
    license="Apache License, Version 2.0",
    version=VERSION,
    packages=["dogsheep_beta"],
    package_data={"dogsheep_beta": ["templates/*.html"]},
    entry_points={
        "datasette": ["beta = dogsheep_beta"],
        "console_scripts": ["dogsheep-beta = dogsheep_beta.cli:cli"],
    },
    install_requires=["datasette", "click", "PyYAML", "sqlite-utils"],
    extras_require={
        "test": ["pytest", "pytest-asyncio", "httpx", "beautifulsoup4", "html5lib"]
    },
    tests_require=["dogsheep-beta[test]"],
)



from datasette import hookimpl
from dogsheep_beta.utils import parse_metadata
import html
import urllib
from jinja2 import Template
import json

TIMELINE_SQL = """
select
  search_index.rowid,
  search_index.[table],
  search_index.key,
  search_index.title,
  search_index.category,
  search_index.timestamp,
  search_index.search_1
from
  search_index
{where}
  {where_clauses}
order by
  search_index.timestamp desc
limit 100
"""

SEARCH_SQL = """
select
  search_index_fts.rank,
  search_index.rowid,
  search_index.[table],
  search_index.key,
  search_index.title,
  search_index.category,
  search_index.timestamp,
  search_index.search_1
from
  search_index join search_index_fts on search_index.rowid = search_index_fts.rowid
{where}
  {where_clauses}
order by
  search_index_fts.rank, search_index.timestamp desc
limit 100
"""
FILTER_COLS = ("table", "category", "is_public")


async def beta(request, datasette):
    from datasette.utils.asgi import Response

    config = datasette.plugin_config("dogsheep-beta") or {}
    database_name = config.get("database") or datasette.get_database().name
    dogsheep_beta_config_file = config["config_file"]
    rules = parse_metadata(open(dogsheep_beta_config_file).read())
    q = request.args.get("q") or ""
    results = []
    facets = {}
    count = None

    results = await search(datasette, database_name, request)
    count, facets = await get_count_and_facets(datasette, database_name, request)
    await process_results(datasette, results, rules)

    hiddens = [
        {"name": column, "value": request.args[column]}
        for column in FILTER_COLS
        if column in request.args
    ]
    return Response.html(
        await datasette.render_template(
            "beta.html",
            {
                "q": q or "",
                "count": count,
                "results": results,
                "facets": facets,
                "hiddens": hiddens,
            },
            request=request,
        )
    )


async def search(datasette, database_name, request):
    database = datasette.get_database(database_name)
    q = request.args.get("q") or ""
    params = {"query": q}
    where_clauses = []
    sql = TIMELINE_SQL
    if q:
        sql = SEARCH_SQL
        where_clauses.append("search_index_fts match :query ")
    for arg in FILTER_COLS:
        if arg in request.args:
            where_clauses.append("[{arg}]=:{arg}".format(arg=arg))
            params[arg] = request.args[arg]
    sql_to_execute = sql.format(
        where=" where " if where_clauses else "",
        where_clauses=" and ".join(where_clauses),
    )
    results = await database.execute(sql_to_execute, params)
    return [dict(r) for r in results.rows]


async def process_results(datasette, results, rules):
    # Adds a 'display' property with HTML to the results
    compiled = {}
    rules_by_table = {}
    for db_name, tables in rules.items():
        for table, meta in tables.items():
            rules_by_table["{}/{}".format(db_name, table)] = meta

    for result in results:
        table = result["table"]
        meta = rules_by_table[table]
        if meta.get("display_sql"):
            db = datasette.get_database(table.split(".")[0])
            display_results = await db.execute(
                meta["display_sql"], {"key": result["key"]}
            )
            first = display_results.first()
            if first:
                result["display"] = dict(first)
            else:
                result["display"] = {}
        output = None
        if meta.get("display"):
            if table not in compiled:
                compiled[table] = Template(meta["display"])
            output = compiled[table].render(**result)
        else:
            output = "<pre>{}</pre>".format(
                html.escape(json.dumps(result, default=repr, indent=4))
            )
        result["output"] = output


async def get_count_and_facets(datasette, database_name, request):
    from datasette.views.table import TableView
    from datasette.utils.asgi import Request, Response

    q = request.args.get("q") or ""
    args = {
        "_facet": ["table", "category", "is_public"],
        "_size": 0,
    }
    if q:
        args["_search"] = q
    for column in FILTER_COLS:
        if column in request.args:
            args[column] = request.args[column]

    path_with_query_string = "/{}/search_index.json?{}".format(
        database_name,
        urllib.parse.urlencode(
            args,
            doseq=True,
        ),
    )
    request = Request.fake(path_with_query_string)
    view = TableView(datasette)
    data, _, _ = await view.data(
        request, database=database_name, hash=None, table="search_index", _next=None
    )
    count, facets = data["filtered_table_rows_count"], data["facet_results"]
    facets = facets.values()
    # Rewrite toggle_url on facet_results
    # ../search_index.json?_search=wolf&_facet=table&_facet=category&_facet=is_public&_size=0&category=2
    for facet in facets:
        for result in facet["results"]:
            bits = urllib.parse.urlparse(result["toggle_url"])
            qs_bits = dict(urllib.parse.parse_qsl(bits.query))
            to_remove = [k for k in qs_bits if k.startswith("_")]
            for k in to_remove:
                qs_bits.pop(k)
            qs_bits["q"] = q
            result["toggle_url"] = "?" + urllib.parse.urlencode(qs_bits)
    return count, facets


@hookimpl
def register_routes():
    return [("/-/beta", beta)]


@hookimpl
def extra_template_vars():
    return {"intcomma": lambda s: "{:,}".format(int(s))}
